/*
 * detector.cpp
 *
 *  Created on: Mar 5, 2012
 *      Author: azfar
 *
 * Referenced the following code: http://scien.stanford.edu/pages/labsite/2001/ee368/projects2001/dropbox/project16/appendix.html
 */

#include "detector.h"

detector::detector() {
}

detector::detector(string t) {
}

detector::~detector() {
	// TODO Auto-generated destructor stub
}

//////	HELPER INTERNAL METHODS	//////

CvHaarClassifierCascade* load_object_detector( const char* cascade_path )
{
    return (CvHaarClassifierCascade*)cvLoad( cascade_path );
}

/*
 * This method is used to crop out everything that is not a face in the image
 * This is done by using haarcascades that were supplied by OpenCV. Make sure that the file
 * "haarcascade_frontalface_alt.xml" exists where this file resides in
 *
 * The code is adapted from online examples for face detection.
 *
 * Note: This needs to be more developed and sometimes fails to find a face.
 */
Mat FaceDetection(string file, string name)
{
	IplImage* image;
	image = cvLoadImage (file.c_str(), 1);
	string cascade_path = "haarcascade_frontalface_alt.xml";

	CvHaarClassifierCascade* cascade = load_object_detector(cascade_path.c_str());

	CvMemStorage* storage = cvCreateMemStorage(0);
	CvSeq* faces;
	int scale = 2.5;
	IplImage* small_image = image;

	small_image = cvCreateImage( cvSize(image->width/2,image->height/2), IPL_DEPTH_8U, 3 );
	cvPyrDown( image, small_image, CV_GAUSSIAN_5x5 );

	/* use the fastest variant */
	faces = cvHaarDetectObjects( small_image, cascade, storage, 1.2, 2, CV_HAAR_DO_CANNY_PRUNING );


	if (faces->total == 0)
	{

		Mat temp(small_image);
		cout<<"No face found for "<<file<<endl;
		return temp;
	}

	/* extract the rectangles only */
	CvRect face_rect = *(CvRect*)cvGetSeqElem( faces, 0 );

    cvSetImageROI(image, cvRect(face_rect.x*scale,face_rect.y*scale,
									(face_rect.width)*scale,(face_rect.height)*scale));

	IplImage *img2 = cvCreateImage(cvGetSize(image),
									   image->depth,
									   image->nChannels);
	cvCopy(image, img2, NULL);
	Mat temp(img2);



	cvResetImageROI(image);
	cvReleaseMemStorage( &storage );
	cvReleaseHaarClassifierCascade( &cascade );


	return temp;


}

/*
 * Method used to convert an integer to a string
 */
string convertInt(int num)
{
	stringstream ss;
	ss << num;
	return ss.str();
}

/*
 * Method used to convert a string to an integer
 */
double convertString(String s)
{
	stringstream ss(s);
	double i;
	ss >> i;
	return i;
}

/*
 * Method used to check if a file exists with the provided filename
 */
bool FileExists(const char *filename)
{
	ifstream ifile(filename);
	return ifile;
}

/*
 * Method calculates the mean of the provided vector
 */
double VectorMean(vector<double> vec)
{
	double sum = 0;

	for(int i = 0; i < vec.size(); i++)
		sum = sum + vec[i];

	return (double(sum)/double(vec.size()));
}


//////	LEARNING ALGORITHMS METHODS	//////

/*
 * This method conducts a simple linear discrimenant technique to train the learning algorithm.
 * This is not used anymore, but can replace the SVM calls within TrainDCT and Classify if wanted.
 * The result of this method are two matrices: test_class and train_class which contain the predicted
 * values from the algorithm
 *
 * The inputs are:
 * 		train_group:	Matrix containing the training classification
 * 		train_data:		Matrix containing the training image information
 * 		test_data:		Matrix containing the test image information
 * 		test_class:		Matrix containing the predicted test classes
 * 		train_class:	Matrix containing the predicted train classes
 */
void LinearDiscriminant(Mat train_group, Mat train_data, Mat test_data, Mat &test_class, Mat &train_class)
{
	// Linear Discriminent
	Mat X,Y, W;

	Y = train_group.clone();
	Y = Y - 1;

	invert(train_data, X, DECOMP_SVD);
	W = Y*X;

	Mat temp;
	temp = W*test_data;

	test_class.create(temp.rows,temp.cols,temp.type());

	for(int i = 0; i < temp.cols; i++)
	{
		if(temp.at<double>(0,i)>0.5)
			test_class.at<double>(0,i) = 2;
		else //if(temp.at<double>(0,i)<=0.5)
			test_class.at<double>(0,i) = 1;
	}

	temp.empty();
	temp = W*train_data;

	train_class.create(temp.rows,temp.cols,temp.type());

	for(int i = 0; i < temp.cols; i++)
	{
		if(temp.at<double>(0,i)>0.5)
			train_class.at<double>(0,i) = 2;
		else //if(temp.at<double>(0,i)<=0.5)
			train_class.at<double>(0,i) = 1;
	}
}

/*
 * This method trains a SVM (Support Vector Machine) based on the passed in data.
 * TrainDCT method calls this function to determine the best SVM. Parts of this code is implemented in
 * the Classify method.
 * The result of this method are two matrices: test_class and train_class which contain the predicted
 * values from the algorithm
 *
 * The inputs are:
 * 		train_group:	Matrix containing the training classification
 * 		train_data:		Matrix containing the training image information
 * 		test_data:		Matrix containing the test image information
 * 		test_class:		Matrix containing the predicted test classes
 * 		train_class:	Matrix containing the predicted train classes
 * 		setNum:			Integer that stores what interation is being ran
 * 		isTrain:		Flag used to know which xml to override
 */
void SupportVector(Mat train_group, Mat train_data, Mat test_data, Mat &test_class, Mat &train_class, int setNum, bool isTrain)
{
	// NOTES: Might need to change to CV_32F1 if float doesnt work

	//Transpose train_group from (0,class) to (class,0)
	Mat trans_train_group;
	transpose(train_group,trans_train_group);
	trans_train_group.convertTo(trans_train_group,DataType<float>::type);
	trans_train_group = trans_train_group - 1;

	//Transpose train_data, test_data from (feature,instance) to (instance,feature)
	Mat trans_train_data;
	transpose(train_data,trans_train_data);
	trans_train_data.convertTo(trans_train_data,DataType<float>::type);

	Mat trans_test_data;
	transpose(test_data,trans_test_data);
	trans_test_data.convertTo(trans_test_data,DataType<float>::type);


	// SVM Parameters
	CvSVMParams params;
	params.svm_type = CvSVM::NU_SVC;
	params.nu = .05;
	params.kernel_type = CvSVM::RBF;
	params.term_crit = cvTermCriteria(CV_TERMCRIT_EPS, 1000, 1e-6);

	CvSVM SVM;

	//Check if a SVM xml exists for the set. If so, load it. Else, train
	if (isTrain == true)
	{
		// Trains the SVM and saves the SVM with the setNum value
		SVM.train_auto(trans_train_data, trans_train_group, Mat(), Mat(), params);
		SVM.save(("Trained/Trained_SVM"+convertInt(setNum)+".xml").c_str());
	}
	else
	{
		// Checks to see if the XML exists
		if(FileExists("Trained/Trained_SVM.xml"))
		{
			SVM.load("Trained/Trained_SVM.xml");
		}
		else
		{
			// Trains the SVM and saves the SVM with the setNum value
			SVM.train_auto(trans_train_data, trans_train_group, Mat(), Mat(), params);
			SVM.save(("Trained/Trained_SVM"+convertInt(setNum)+".xml").c_str());
		}
	}

	// Predict and store train results
	train_class.create(train_group.rows,train_group.cols,train_group.type());
	for(int i = 0; i < trans_train_data.rows; i++)
	{
			Mat testSample = trans_train_data.row(i);
			double response = (double)SVM.predict(testSample);
			train_class.at<double>(0,i) = response+1;
	}


	// Predict and store test results
	test_class.create(test_data.rows,test_data.cols,test_data.type());
	for(int i = 0; i < trans_test_data.rows; i++)
	{
			Mat testSample = trans_test_data.row(i);
			double response = (double)SVM.predict(testSample);
			test_class.at<double>(0,i) = response+1;
	}


}

/*
 * This method is very similar to the SupportVector method.
 * TrainDCT_Stats method calls this function to determine the best SVM based on various statistics.
 * The result of this method are two matrices: test_class and train_class which contain the predicted distance
 * values from the algorithm which are used to calculate True Positive, True Negative, etc.
 *
 * The inputs are:
 * 		train_group:	Matrix containing the training classification
 * 		train_data:		Matrix containing the training image information
 * 		test_data:		Matrix containing the test image information
 * 		test_class:		Matrix containing the predicted test distance
 * 		train_class:	Matrix containing the predicted train distamce
 * 		setNum:			Integer that stores what interation is being ran
 * 		isTrain:		Flag used to know which xml to override
 */
void SupportVector_Stats(Mat train_group, Mat train_data, Mat test_data, Mat &test_class, Mat &train_class, int setNum, bool isTrain)
{
	// NOTES Might need to change to CV_32F1

	//Transpose train_group from (0,class) to (class,0)
	Mat trans_train_group;
	transpose(train_group,trans_train_group);
	trans_train_group.convertTo(trans_train_group,DataType<float>::type);
	trans_train_group = trans_train_group - 1;

	//Transpose train_data, test_data from (feature,instance) to (instance,feature)
	Mat trans_train_data;
	transpose(train_data,trans_train_data);
	trans_train_data.convertTo(trans_train_data,DataType<float>::type);

	Mat trans_test_data;
	transpose(test_data,trans_test_data);
	trans_test_data.convertTo(trans_test_data,DataType<float>::type);


	// SVM Parameters
	CvSVMParams params;
	params.svm_type = CvSVM::C_SVC;
	params.nu = .01;
	params.kernel_type = CvSVM::RBF;
	params.term_crit = cvTermCriteria(CV_TERMCRIT_EPS, 1000, 1e-6);

	CvSVM SVM;

	//Check if a SVM xml exits for the set. If so, load it. Else, train
	if (isTrain == true)
	{
		SVM.train(trans_train_data, trans_train_group, Mat(), Mat(), params);
		SVM.save(("Trained/Trained_SVM"+convertInt(setNum)+".xml").c_str());
	}
	else
	{
		if(FileExists("Trained/Trained_SVM.xml"))
		{
			SVM.load("Trained/Trained_SVM.xml");
		}
		else
		{
			SVM.train_auto(trans_train_data, trans_train_group, Mat(), Mat(), params);
			SVM.save(("Trained/Trained_SVM"+convertInt(setNum)+".xml").c_str());
		}
	}

	// Predict and store test results
	train_class.create(train_group.rows,train_group.cols,train_group.type());
	for(int i = 0; i < trans_train_data.rows; i++)
	{
			Mat testSample = trans_train_data.row(i);
			double dist = (double)SVM.predict(testSample, true );	// This call returns the distance instead of a class
			train_class.at<double>(0,i) = dist;
	}

	// Predict and store test results
	test_class.create(test_data.rows,test_data.cols,test_data.type());
	for(int i = 0; i < trans_test_data.rows; i++)
	{
			Mat testSample = trans_test_data.row(i);
			double dist = (double)SVM.predict(testSample, true );	// This call returns the distance instead of a class
			test_class.at<double>(0,i) = dist;
	}

}



//////	DISCRETE COSINE TRANSFORM METHODS	//////

/*
 * This method loads the DCT coefficients used for image processing.
 * The return value is a matrix that contains a list of coefficients.
 * The file read is Trained/DCT_Coef.dat
 *
 * The inputs are:
 * 		dctCoefs:	Matrix containing the coefficients
 * 		N:			The number of coefficients
 */
void LoadDCTCoef(Mat &dctCoefs, int &N)
{
	ifstream inFile("Trained/DCT_Coef.dat");
	String line;

	if(inFile.is_open())
	{
		// Load Data
		getline(inFile, line);
		N = convertString(line);
		getline(inFile, line);

		dctCoefs.create(N, 1, DataType<int>::type);

		int x = 0;
		while (inFile.good())
		{
			getline(inFile,line);
			dctCoefs.at<int>(x,0) = convertString(line);
			x++;
		}
	}
}

/*
 * This method saves the DCT coefficients that are used for image processing.
 * The file saved is Trained/DCT_Coef.dat
 *
 * The inputs are:
 * 		dctCoefs:	Matrix containing the coefficients
 */
void SaveDCTCoef(Mat &dctCoefs)
{
	ofstream outFile("Trained/DCT_Coef.dat");
	String line;

	outFile<<dctCoefs.rows<<endl;	// Stores the number of coefficients in the first line
	outFile<<endl;

	for(int i =0; i< dctCoefs.rows; i++)
		outFile<<convertInt(dctCoefs.at<int>(i,0))<<endl;

	outFile.close();
}

/*
 * This method determines the top N coefficients that will be used for image processing.
 * The OpenCV dct method is used to calculate the DCT coefficients. This method runs through
 * all the training set by each column (column of pixels) and does the following
 * 		1) Calculates the DCT values with zero means
 * 		2) Store all the dct coefficient sums
 * 		3) Determine the top N coefficients
 *
 * The return value is a matrix that containing the top N coefficients.
 *
 * The inputs are:
 * 		dctCoefs:	Matrix containing the coefficient indices
 * 		trainSet:	DCT values for each image in the train set
 * 		size:		Size of the images (pixels)
 * 		N:			The number of coefficients
 */
void DetermineDCTCoefIdx(Mat &dctCoefs, Mat trainSet, int size, int N)
{
	Mat dct_coef_sum((size*size),1,  DataType<float>::type, 0);

	// Sum DCT coefficients for all images in trainSet by columns
	for(int i = 0; i<trainSet.cols; i++)
	{
		Mat img_vec;
		trainSet.col(i).copyTo(img_vec);

		img_vec.convertTo(img_vec, DataType<float>::type);

		// Subtract out the mean so all DCT have zero mean images
		img_vec = img_vec - mean(img_vec);

		Mat dct_img;
		dct(img_vec, dct_img);		// Calls the OpenCV DCT method to get the values and coefficients

		// Maintain total of DCT coefficient magnitudes
		dct_coef_sum += abs(dct_img);
	}

	// maxCoef contains indices of the top N coefficients
	Mat maxCoef;
	sortIdx(dct_coef_sum, maxCoef, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
	maxCoef.convertTo(maxCoef,DataType<int>::type);

	//Convert maxCoefs to be row aligned with top N values
	maxCoef.rowRange(Range(0,N)).copyTo(dctCoefs);
}

/*
 * This method takes in a matrix of all the images and calculates their respective DCT values
 * using the coefficients found in DetermineDCTCoefIdx.
 * The OpenCV dct method is called with the DCT_ROWS flag to indicate each row is a different image.
 * For more detail on the dct method, look into the OpenCV manual.
 *
 * The inputs are:
 * 		dataMat:	Matrix containing the image raw data
 * 		dctDataMat:	Matrix containing the resulting DCT data for each image
 * 		dctCoefs:	Matrix containing the top N coefficients to be used
 *
 */
void ConvertToDCT(Mat dataMat, Mat &dctDataMat, Mat dctCoefs)
{
	// Store the top indices from the total set
	for(int i = 0; i < dataMat.cols; i++ )
		for(int x = 0; x < dctCoefs.rows; x++)
			dctDataMat.at<double>(x,i) = dataMat.at<double>(dctCoefs.at<int>(x,0),i);

	Mat dctDataMat_t, dctDataMatConvert;
	transpose(dctDataMat, dctDataMat_t);
	dctDataMat_t.convertTo(dctDataMat_t, DataType<float>::type);

	dct(dctDataMat_t, dctDataMatConvert, DCT_ROWS);		// Calls the OpenCV DCT method indicating each row is a seperate image
	transpose(dctDataMatConvert, dctDataMat);
	dctDataMat.convertTo(dctDataMat, DataType<double>::type);
}



//////	MAIN DETECTOR METHODS	//////

/*
 * This method loads the face images from the dirname directory. The directory must contain only images and two
 * folders: Positive and Negative. The result of the method is facesMatND which is a three dimensional array that
 * contains intesity image information in the format of [x-pixel][y-pixel][image] of type double
 *
 * The inputs are:
 * 		dirname:	Directory containing Positive and Negative images
 * 		size:		Size of the images to be reduced to (images set as size x size)
 * 		facesMatND:	Three dimensional array containing pixel information about each image
 * 		Num_Pos:	Number of positive images
 * 		Num_Neg:	Number of negative images
 * 		detectFace:	Flag used to turn on face detection algorithm to focus only on the face
 */
void detector::LoadFaces (string dirname, int size, MatND &facesMatND, int Num_Pos, int Num_Neg, bool detectFace)
{

	cout<<"Starting to Load Faces"<<endl;

	DIR *pDIR;
	struct dirent *entry;

	string pfiles [Num_Pos];
	string nfiles [Num_Neg];

	int i=0;

	// Store image names into positive and negative arrays

	// Read in positive filenames
	if( pDIR=opendir((dirname+"/Positive").c_str()) )
	{
		while((entry = readdir(pDIR))&&(i<Num_Pos))
		{
			if( strcmp(entry->d_name, ".") != 0 && strcmp(entry->d_name, "..") != 0 )
			{
				pfiles[i] = entry->d_name;
				i++;
			}
		}
		closedir(pDIR);
	}

	i=0;
	// Read in negative filenames
	if( pDIR=opendir((dirname+"/Negative").c_str()) )
	{
		while((entry = readdir(pDIR))&&(i<Num_Neg))
		{
			if( strcmp(entry->d_name, ".") != 0 && strcmp(entry->d_name, "..") != 0 )
			{
				nfiles[i] = entry->d_name;
				i++;
			}
		}
		closedir(pDIR);
	}

	// Initialize space for the faces.
	int sy = size;
	int sx = size;
	int sz = Num_Pos+Num_Neg;

	// Load the Positive images and reduce them to the size value utilizing
	// nearest-neighbor interpolation
	// Each image is stored into facesMatND with a normalized greyscale value [0,1]
	for(int i = 0; i< Num_Pos; i++)
	{
		Mat image;

		// Checks to see if face detection is required
		if(detectFace == false)
			image = imread((dirname+"/Positive/"+pfiles[i]).c_str(),0);
		else
			image = FaceDetection((dirname+"/Positive/"+pfiles[i]).c_str(), pfiles[i]);

		if(image.cols != 1)
		{
			Mat img_resized(sx,sy,CV_8UC1);

			// Resizes the image to sx,sy and greyscale
			resize (image,img_resized,img_resized.size(),0,0,INTER_NEAREST);

			for(int x = 0; x<sx; x++)
				for(int y = 0; y<sy; y++)
				{
					// Get the greyscale value at each pixel
					Scalar intensity = img_resized.at<uchar>(x,y);
					facesMatND.at<double>(x,y,i) = intensity[0]/255;
				}
		}

	}

	// Load the Negative images and reduce them to the size value utilizing nearest-neighbor interpolation
	// Each image is stored into facesMatND with a normalized greyscale value [0,1]
	// These images are added after the Positive images
	for(int i = 0; i< Num_Neg; i++)
	{
		Mat image;

		// Checks to see if face detection is required
		if(detectFace == false)
			image = imread((dirname+"/Negative/"+nfiles[i]).c_str(),0);
		else
			image = FaceDetection((dirname+"/Negative/"+nfiles[i]).c_str(), nfiles[i]);

		if(image.cols != 1)
		{
			Mat img_resized(sx,sy,CV_8UC1);

			// Resizes the image to sx,sy and greyscale
			resize (image,img_resized,img_resized.size(),0,0,INTER_NEAREST);

			for(int x = 0; x<sx; x++)
				for(int y = 0; y<sy; y++)
				{
					// Get the greyscale value at each pixel
					Scalar intensity = img_resized.at<uchar>(x,y);
					facesMatND.at<double>(x,y,i+Num_Pos) = intensity[0]/255;
				}
		}
	}

	cout<<"Finished Load Faces"<<endl;
}


/*
 * This method trains the classifier using DCT (Discrete Cosine Transform) and SVM (Support Vector Machines).
 * The method takes in the facesMatND and the truth vector, trains the SVM, and stores the best SVM into the folder
 * Trained/Trained_SVM.xml
 *
 * The inputs are:
 * 		size:			Size of the images (images set as size x size)
 * 		facesMatND:		Three dimensional array containing pixel information about each image
 * 		classification:	Vector containing the ground truth (1 = Positive, 2 = Negative)
 * 		Num_Pos:		Number of positive images
 * 		Num_Neg:		Number of negative images
 * 		N:				Number of coefficients used in DCT. 310 usually used
 * 		isTrain:		Flag to indicate if a new SVM needs to be used or the existing one can be used. Used more for retesting
 */
void detector::TrainDCT(int size, MatND &facesMatND, vector<double> classification, int NUM_POS, int NUM_NEG, int N, bool isTrain)
{
	vector<double> actrain, actest;		// Vectors to contain accuracies from training and testing
	double bestVal = 0;					// Value of the best accuracy

	//Convert facesMatND to a Mat of size [size*size][#images]
	Mat facesMat((facesMatND.size[0]*facesMatND.size[1]), facesMatND.size[2], DataType<double>::type);
	for(int i = 0; i < facesMatND.size[2]; i++)
		for(int x = 0; x < facesMatND.size[0]; x++)
			for(int y = 0; y < facesMatND.size[1]; y++)
				facesMat.at<double>(x*y,i) = facesMatND.at<double>(x,y,i);


	vector <int> pos_index;		// Vector containing the positive indices within classification
	vector <int> neg_index;		// Vector containing the negative indices within classification


	for(int i = 0; i<classification.size(); i++)
	{
		if(classification[i] == 1)
			pos_index.push_back(i);
		else
			neg_index.push_back(i);
	}

	// Randomize the indices for testing/training purposes
	srand ( unsigned ( time (NULL) ) );
	std::random_shuffle(pos_index.begin(), pos_index.end());
	std::random_shuffle(neg_index.begin(), neg_index.end());

	int num_test = ((NUM_POS+NUM_NEG)*.3)/2;			// Sets the split of training/testing. Set as 70/30
	cout<< "Number of Test = "<<num_test<<endl;

	// Best N=310
	/* This is the main loop used to train the SVM. There are 10 iterations where randomly picked
	 * images from the training and testing set are chosen for training and testing. The best trained
	 * SVM is saved for future classification
	 */
	for(int m = 1; m < 10; m+=1)
	{
		vector <int> test_index;	// Vector containing the testing indices
		vector <int> train_index;	// Vector containing the training indices

		// Take num_test number of indices as test data
		// For Positive
		for(int i = 0; i < num_test; i++)
			test_index.push_back(pos_index[(i+(m*num_test)) % pos_index.size()]);

		// For Negative
		for(int i = 0; i < num_test; i++)
			test_index.push_back(neg_index[(i+(m*num_test)) % neg_index.size()]);

		// Get the training data which is XOR of the test_data
		for(int i = 0; i < classification.size(); i++)
		{
			bool found = false;
			for (int c = 0; c < test_index.size(); c++)
			{
				if (test_index[c] == i)
					found = true;
			}

			if(found == false)
				train_index.push_back(i);
		}

		// Gets the training image information
		Mat facesTrain(facesMat.rows, train_index.size(), DataType<double>::type);
		for(int i = 0; i < train_index.size(); i++)
				for(int x = 0; x < facesMat.rows; x++)
					facesTrain.at<double>(x,i) = facesMat.at<double>(x,train_index[i]);



		// Determine the most variable DCT coefficients from the training set
		Mat dctCoefs;
		if(isTrain == true)
			DetermineDCTCoefIdx(dctCoefs, facesTrain, size, N);

		// Read DCT Coefs from file
		if(isTrain == false)
			LoadDCTCoef(dctCoefs, N);

		// Convert Data to top N DCT coefficients
		Mat dctFacesMat(N,facesMat.cols,DataType<double>::type);
		ConvertToDCT(facesMat, dctFacesMat, dctCoefs);

		// Create test data
		Mat test_data(N, test_index.size(), DataType<double>::type);	// Contains test image data
		Mat test_group(1,test_index.size(), DataType<double>::type);	// Contains test classification data

		for (int x = 0; x < N; x++)
		{
			int y = 0;
			for (int i = 0; i < test_index.size(); i++)
			{
				//Sets the pixel information for the test data
				test_data.at<double>(x,y) = dctFacesMat.at<double>(x,test_index[i]);
				y++;

				// Sets the label/classification for the test data
				if(x == 0)
					test_group.at<double>(0,i) = classification[test_index[i]];
			}
		}


		// Create train data
		Mat train_data(N, train_index.size(), DataType<double>::type);	// Contains train image data
		Mat train_group(1,train_index.size(), DataType<double>::type);	// Contains train classification data
		for (int x = 0; x < N; x++)
		{
			int y = 0;
			for (int i = 0; i < train_index.size(); i++)
			{
				//Sets the pixel information for the train data
				train_data.at<double>(x,y) = dctFacesMat.at<double>(x,train_index[i]);
				y++;

				// Sets the label/classification for the train data
				if(x == 0)
					train_group.at<double>(0,i) = classification[train_index[i]];
			}
		}

		/* Execute algorithm (SVM)*/
		Mat test_class, train_class;
		SupportVector(train_group, train_data, test_data, test_class, train_class, m, isTrain);

		// Determine Accuracy of training based on training set and test set
		int train_correct = 0, test_correct = 0;
		for(int i = 0; i < train_group.cols; i++)
		{
			if(train_group.at<double>(0,i) == train_class.at<double>(0,i))
				train_correct++;
		}

		for(int i = 0; i < test_group.cols; i++)
		{
			if(test_group.at<double>(0,i) == test_class.at<double>(0,i))
				test_correct++;
		}

		// Store the accuracy for both sets
		actrain.push_back(double(train_correct)/double(train_group.cols));
		actest.push_back(double(test_correct)/double(test_group.cols));

		// If training, the best accuracy Coefs and SVM are saved
		if(isTrain == true)
		{
			if(actest[actest.size()-1] >= bestVal)
			{
				// Saves the Coefficient indices to a file
				SaveDCTCoef(dctCoefs);

				// Save the SVM to Trained folder
				CvSVM SVM;
				SVM.load(("Trained/Trained_SVM"+convertInt(m)+".xml").c_str());
				SVM.save("Trained/Trained_SVM.xml");

				bestVal = actest[actest.size()-1];
				cout<< "\tBest Val = "<<bestVal<<endl;
			}
		}

		cout<<"m = "<<m<<" N = "<<N<<" Correct: "<<actrain[actrain.size()-1]<<" "<<actest[actest.size()-1]<<endl;
	}

	double train_acc = VectorMean(actrain);
	double test_acc =  VectorMean(actest);
	cout<<"Average <Training, Testing>: < "<<train_acc<<",\t"<<test_acc<<" >"<<endl;
}

/*
 * This method works in the same fashion as TrainDCT except it outputs additional statistics including
 * True Positive, True Negative, False Positive, False Negative. This method is used for ROC curve generation.
 * The theta value is varied between -2 and 2 and various data is printed to analyze the performance
 *
 * Refer the TrainDCT for details
 */
void detector::TrainDCT_Stats(int size, MatND &facesMatND, vector<double> classification, int NUM_POS, int NUM_NEG, int N, bool isTrain)
{
	vector<double> actrain, actest;

	//Convert facesMatND to a Mat of size [size*size][#images]
	Mat facesMat((facesMatND.size[0]*facesMatND.size[1]), facesMatND.size[2], DataType<double>::type);

	for(int i = 0; i < facesMatND.size[2]; i++)
		for(int x = 0; x < facesMatND.size[0]; x++)
			for(int y = 0; y < facesMatND.size[1]; y++)
				facesMat.at<double>(x*y,i) = facesMatND.at<double>(x,y,i);

	vector <int> pos_index;
	vector <int> neg_index;
	double bestVal = 0;

	for(int i = 0; i<classification.size(); i++)
	{
		if(classification[i] == 1)
			pos_index.push_back(i);
		else
			neg_index.push_back(i);
	}

	// Randomize the indices for testing/training purposes
	srand ( unsigned ( time (NULL) ) );
	std::random_shuffle(pos_index.begin(), pos_index.end());
	std::random_shuffle(neg_index.begin(), neg_index.end());

	// Loops for varying theta values which are used within SVM to classify
	for(double theta = -2; theta < 2; theta+=.2)
	{

		int num_test = ((NUM_POS+NUM_NEG)*.3)/2;
		cout<< "Number of Test = "<<num_test<<endl;

		// Various vectors used for performance
		double bestVal = 0;
		vector<double> actrain, actest;
		vector<double> TPtrainVec, FPtrainVec, FNtrainVec, TNtrainVec;
		vector<double> TPtestVec, FPtestVec, FNtestVec, TNtestVec;

		for(int m = 1; m < 10; m+=1)
		{
			int N = 100; // N = 100 is used to allow better results

			vector <int> test_index;
			vector <int> train_index;

			// Take num_test number of indices as test data
			// For Positive
			for(int i = 0; i < num_test; i++)
				test_index.push_back(pos_index[(i+(m*num_test)) % pos_index.size()]);

			// For Negative
			for(int i = 0; i < num_test; i++)
				test_index.push_back(neg_index[(i+(m*num_test)) % neg_index.size()]);

			// Get the training data which is XOR of the test_data
			for(int i = 0; i < classification.size(); i++)
			{
				bool found = false;
				for (int c = 0; c < test_index.size(); c++)
				{
					if (test_index[c] == i)
						found = true;
				}

				if(found == false)
					train_index.push_back(i);
			}


			Mat facesTrain(facesMat.rows, train_index.size(), DataType<double>::type);
			for(int i = 0; i < train_index.size(); i++)
					for(int x = 0; x < facesMat.rows; x++)
						facesTrain.at<double>(x,i) = facesMat.at<double>(x,train_index[i]);


			Mat dctCoefs;
			// Determine the most variable DCT coefficients from the training set
			if(isTrain == true)
				DetermineDCTCoefIdx(dctCoefs, facesTrain, size, N);

			// Read DCT Coefs from file
			if(isTrain == false)
				LoadDCTCoef(dctCoefs, N);

			// Convert Data to top N DCT coefficients
			Mat dctFacesMat(N,facesMat.cols,DataType<double>::type);
			ConvertToDCT(facesMat, dctFacesMat, dctCoefs);

			// Create test data
			Mat test_data(N, test_index.size(), DataType<double>::type);
			Mat test_group(1,test_index.size(), DataType<double>::type);

			for (int x = 0; x < N; x++)
			{
				int y = 0;
				for (int i = 0; i < test_index.size(); i++)
				{
					test_data.at<double>(x,y) = dctFacesMat.at<double>(x,test_index[i]);
					y++;

					if(x == 0)
						test_group.at<double>(0,i) = classification[test_index[i]];
				}
			}


			// Create train data
			Mat train_data(N, train_index.size(), DataType<double>::type);
			Mat train_group(1,train_index.size(), DataType<double>::type);
			for (int x = 0; x < N; x++)
			{
				int y = 0;
				for (int i = 0; i < train_index.size(); i++)
				{
					train_data.at<double>(x,y) = dctFacesMat.at<double>(x,train_index[i]);
					y++;

					if(x == 0)
						train_group.at<double>(0,i) = classification[train_index[i]];
				}
			}

			/* Execute algorithm (SVM)
			 * SupportVector_Stats is called instead of the normal SupportVector method.
			 * This method returns the distance from the support vectors which is compared to theta
			 */
			Mat test_class, train_class;
			SupportVector_Stats(train_group, train_data, test_data, test_class, train_class, m, isTrain);

			// TP = True Positive, FP = False Positive, FN = False Negative, TN = True Negative
			// Y0 = Number of class 2, Y1 = Number of class 1
			int TP_train = 0, FP_train = 0, FN_train = 0, TN_train = 0, Y0_train = 0, Y1_train = 0;
			int TP_test = 0, FP_test = 0, FN_test = 0, TN_test = 0, Y0_test = 0, Y1_test = 0;

			// Determine Accuracy of training
			int train_correct = 0, test_correct = 0;
			for(int i = 0; i < train_class.cols; i++)
			{
				// Manually determine the output class by looking at the theta value
				if(train_class.at<double>(0,i) > theta)
					train_class.at<double>(0,i) = 1;
				else
					train_class.at<double>(0,i) = 2;

				// Determine which bucket the results fall into
				if(train_group.at<double>(0,i) == train_class.at<double>(0,i))
					train_correct++;
				if((train_group.at<double>(0,i) == 2) && (train_class.at<double>(0,i) == 2))
					TP_train++;
				if((train_group.at<double>(0,i) == 1) && (train_class.at<double>(0,i) == 2))
					FP_train++;
				if((train_group.at<double>(0,i) == 2) && (train_class.at<double>(0,i) == 1))
					FN_train++;
				if((train_group.at<double>(0,i) == 1) && (train_class.at<double>(0,i) == 1))
					TN_train++;

				if(train_group.at<double>(0,i) == 2)
					Y1_train++;
				if(train_group.at<double>(0,i) == 1)
					Y0_train++;
			}

			// Determine Accuracy of testing
			for(int i = 0; i < test_group.cols; i++)
			{
				// Manually determine the output class by looking at the theta value
				if(test_class.at<double>(0,i) > theta)
					test_class.at<double>(0,i) = 1;
				else
					test_class.at<double>(0,i) = 2;

				// Determine which bucket the results fall into
				if(test_group.at<double>(0,i) == test_class.at<double>(0,i))
					test_correct++;
				if((test_group.at<double>(0,i) == 2) && (test_class.at<double>(0,i) == 2))
					TP_test++;
				if((test_group.at<double>(0,i) == 1) && (test_class.at<double>(0,i) == 2))
					FP_test++;
				if((test_group.at<double>(0,i) == 2) && (test_class.at<double>(0,i) == 1))
					FN_test++;
				if((test_group.at<double>(0,i) == 1) && (test_class.at<double>(0,i) == 1))
					TN_test++;

				if(test_group.at<double>(0,i) == 2)
					Y1_test++;
				if(test_group.at<double>(0,i) == 1)
					Y0_test++;
			}

			//Store and calculate the appropriate rates
			TPtrainVec.push_back(double(TP_train)/double(Y1_train));
			FPtrainVec.push_back(double(FP_train)/double(Y0_train));
			FNtrainVec.push_back(double(FN_train)/double(Y1_train));
			TNtrainVec.push_back(double(TN_train)/double(Y0_train));

			TPtestVec.push_back(double(TP_test)/double(Y1_test));
			FPtestVec.push_back(double(FP_test)/double(Y0_test));
			FNtestVec.push_back(double(FN_test)/double(Y1_test));
			TNtestVec.push_back(double(TN_test)/double(Y0_test));

			actrain.push_back(double(train_correct)/double(train_group.cols));
			actest.push_back(double(test_correct)/double(test_group.cols));

			cout<<"m = "<<m<<" theta = "<<theta<<" TRAIN: TP = "<<TP_train<<" FP = "<<FP_train<<" FN = "<<FN_train<<" TN = "<<TN_train<<" Y1 = "<<Y1_train<<" Y0 = "<<Y0_train<<endl;
			cout<<"\t\t"<<" TEST: TP = "<<TP_test<<" FP = "<<FP_test<<" FN = "<<FN_test<<" TN = "<<TN_test<<" Y1 = "<<Y1_test<<" Y0 = "<<Y0_test<<endl;
		}

		// Calculates the mean value of the vector
		double train_acc = VectorMean(actrain);
		double test_acc =  VectorMean(actest);

		// Calculate the mean value of all the vectors
		cout<<"\t\ttheta = "<<theta<<" Average <Training, Testing>: < "<<train_acc<<",\t"<<test_acc<<" >"<<endl;
		double train_TPR = VectorMean(TPtrainVec);
		double train_FPR = VectorMean(FPtrainVec);
		double train_FNR = VectorMean(FNtrainVec);
		double train_TNR = VectorMean(TNtrainVec);
		double test_TPR = VectorMean(TPtestVec);
		double test_FPR = VectorMean(FPtestVec);
		double test_FNR = VectorMean(FNtestVec);
		double test_TNR = VectorMean(TNtestVec);
		cout<<"theta=\t"<<theta<<"\t"<< train_TPR<<"\t"<< train_FPR<<"\t"<< train_FNR<<"\t"<< train_TNR
			<<"\t"<< test_TPR<<"\t"<< test_FPR<<"\t"<< test_FNR<<"\t"<< test_TNR<<endl;

	}
	double train_acc = VectorMean(actrain);
	double test_acc =  VectorMean(actest);
	cout<<"Average <Training, Testing>: < "<<train_acc<<",\t"<<test_acc<<" >"<<endl;
}



/*
 * This method is used to classify images that are stored in facesMatND. The results are stored into a matrix
 * called predict_class. Currently, DCT and SVM are loaded from the Trained/ folder and used to classify. TrainDCT
 * must be ran before this can be used
 *
 * The inputs are:
 * 		facesMatND:		Three dimensional array containing pixel information about each image
 * 		predict_class:	This matrix contains a list of the classification in the order that they are in facesMatND
 */
void detector::Classify(MatND &facesMatND, Mat &predict_class)
{
	//Convert facesMatND to a Mat of size [size*size][#images]
	Mat facesMat((facesMatND.size[0]*facesMatND.size[1]), facesMatND.size[2], DataType<double>::type);
	for(int i = 0; i < facesMatND.size[2]; i++)
		for(int x = 0; x < facesMatND.size[0]; x++)
			for(int y = 0; y < facesMatND.size[1]; y++)
				facesMat.at<double>(x*y,i) = facesMatND.at<double>(x,y,i);


	Mat dctCoefs;	// Matrix used to store the DCT Coefficient indices
	int N;			// Stores the number of DCT Coefficients

	// Read DCT Coefs from file
	LoadDCTCoef(dctCoefs, N);

	// Convert Data to top N DCT coefficients
	Mat dctFacesMat(N,facesMat.cols,DataType<double>::type);
	ConvertToDCT(facesMat, dctFacesMat, dctCoefs);

	/* Execute algorithm */
	Mat trans_test_data;
	transpose(dctFacesMat,trans_test_data);								// Matrix needs to be in row order
	trans_test_data.convertTo(trans_test_data,DataType<float>::type);	// and of type float

	// SVM Parameters
	CvSVMParams params;
	params.svm_type = CvSVM::NU_SVC;
	params.nu = .05;
	params.kernel_type = CvSVM::RBF;
	params.term_crit = cvTermCriteria(CV_TERMCRIT_EPS, 1000, 1e-6);

	CvSVM SVM;

	// Check if the trained SVM exists
	if(FileExists("Trained/Trained_SVM.xml"))
	{
		SVM.load("Trained/Trained_SVM.xml");
	}
	else
	{
		cout<< "ERROR! SVM File Missing!"<<endl;
	}

	// Predict and store test results
	predict_class.create(1,dctFacesMat.cols,dctFacesMat.type());
	for(int i = 0; i < trans_test_data.rows; i++)
	{
		Mat testSample = trans_test_data.row(i);
		double response = (double)SVM.predict(testSample);
		predict_class.at<double>(0,i) = response+1;
	}
}





//////	DEPRECATED CODE FROM KLT/JACKNIFE PROCESS	//////

void LoadFaceData(int num, Mat &weights, vector<int>&train_index, vector<int>&test_index, vector<double> &classification)
{
	ifstream inFile(("TrainedData/face"+convertInt(num)+".dat").c_str());
	String line;

	if(inFile.is_open())
	{
			// Load Weights
			getline(inFile, line);
			getline(inFile, line);
			double rows = convertString(line);
			getline(inFile, line);
			double cols = convertString(line);

			weights.create(rows, cols, DataType<double>::type);

			int x = 0, y = 0;
			getline(inFile,line,' ');
			while (line.compare(0,1,"#") != 0)
			{
				if(line.compare(0,1,"\n") == 0)
				{
					x++;
					y=0;
				}
				else
				{
					weights.at<double>(x,y) = convertString(line);
					y++;
				}
				getline(inFile,line,' ');
			}

			// Load Train_Index
			getline(inFile, line);
			getline(inFile,line);
			getline(inFile,line,' ');
			while (line.compare(0,1,"#") != 0)
			{
				train_index.push_back(convertString(line));
				getline(inFile,line,' ');
			}
			train_index.pop_back();

			// Load Test_Index
			getline(inFile, line);
			getline(inFile,line);
			getline(inFile,line,' ');
			while (line.compare(0,1,"#") != 0)
			{
				test_index.push_back(convertString(line));
				getline(inFile,line,' ');
			}
			test_index.pop_back();

			// Load Classification Information
			getline(inFile, line);
			getline(inFile,line,' ');
			while (line.compare(0,1,"#") != 0)
			{
				//cout<<line<<endl;
				classification.push_back(convertString(line));
				getline(inFile,line,' ');
			}
			classification.pop_back();

		inFile.close();
	}

}

void SaveFaceData(int num, Mat weight, vector<int>train_index, vector<int>test_index, vector<double> classification)
{

	ofstream outFile;
	outFile.open(("TrainedData/face"+convertInt(num)+".dat").c_str());

	// Writes the weight data into the file
	outFile <<"##### WEIGHTS #####\n";
	outFile <<weight.rows<<"\n"<<weight.cols<<"\n";
	for(int i = 0; i < weight.rows; i++)
	{
		for(int j = 0; j < weight.cols; j++)
			outFile << weight.at<double>(i,j)<<" ";

		outFile << "\n ";
	}

	// Writes the train_index data to the file
	outFile <<"##### TRAIN_INDEX #####\n";
	outFile <<train_index.size()<<"\n";
	for(int i = 0; i < train_index.size(); i++)
	{
		outFile << train_index[i]<<" ";
	}
	outFile << "\n ";

	// Writes the train_index data to the file
	outFile <<"##### TEST_INDEX #####\n";
	outFile <<test_index.size()<<"\n";
	for(int i = 0; i < test_index.size(); i++)
	{
		outFile << test_index[i]<<" ";
	}
	outFile << "\n ";

	// Writes the classification data to the file
	outFile <<"##### CLASSIFICATION #####\n";
	for(int i = 0; i < classification.size(); i++)
	{
		outFile << classification[i]<<" ";
	}
	outFile << "\n # ";

	outFile.close();
}

void JackKnife(int size, MatND &facesMatND, vector<double> classification)
{
	cout<<"Starting JackKnife"<<endl;

	srand(time(NULL));

	//Convert facesMatND to a Mat of size [size*size][#images]
	Mat facesMat((facesMatND.size[0]*facesMatND.size[1]), facesMatND.size[2], DataType<double>::type);

	for(int i = 0; i < facesMatND.size[2]; i++)
		for(int x = 0; x < facesMatND.size[0]; x++)
			for(int y = 0; y < facesMatND.size[1]; y++)
				facesMat.at<double>(x*y,i) = facesMatND.at<double>(x,y,i);

	vector <int> pos_index;
	vector <int> neg_index;

	for(int i = 0; i<classification.size(); i++)
	{
		if(classification[i] == 1)
			pos_index.push_back(i);
		else
			neg_index.push_back(i);
	}

	// Randomize the indices for testing/training purposes
	srand ( unsigned ( time (NULL) ) );
	std::random_shuffle(pos_index.begin(), pos_index.end());
	std::random_shuffle(neg_index.begin(), neg_index.end());

	int num_test = 20;

	for(int m = 0; m < 10; m++)
	{
		cout<< "Started M = "<< (m+1) <<endl;

		vector <int> test_index;
		vector <int> train_index;

		// Take num_test number of indices as test data
		// For Positive
		for(int i = 0; i < num_test; i++)
			test_index.push_back(pos_index[i+(m*num_test)]);

		// For Negative
		for(int i = 0; i < num_test; i++)
			test_index.push_back(neg_index[i+(m*num_test)]);

		// Get the training data which is XOR of the test_data
		for(int i = 0; i < classification.size(); i++)
		{
			bool found = false;
			for (int c = 0; c < test_index.size(); c++)
			{
				if (test_index[c] == i)
					found = true;
			}

			if(found == false)
				train_index.push_back(i);
		}

		Mat facesTrain(facesMat.rows, train_index.size(), DataType<double>::type);
		for(int i = 0; i < train_index.size(); i++)
				for(int x = 0; x < facesMat.rows; x++)
					facesTrain.at<double>(x,i) = facesMat.at<double>(x,train_index[i]);

		// Multiply training set with its transpose
		Mat iauto;
		iauto = facesTrain.t()*facesTrain;

		Mat lamda, vec;
		eigen(iauto, lamda, vec);
		//lamda = lamda.diag(0);  Commented out since lamda already contains the main diagonal

		//NOTE: Orignal matlab code reversed vec and lamda. Since lamda is sorted already,
		// 		it is assumed that vec is as well

		Mat eig_faces;
		eig_faces = facesTrain*vec;

		// This is the equivalent of: eig_faces = eig_faces .*repmat( 1./sqrt(eig_values'), [size(eig_faces,1) 1]);
		Mat lamda_sqrt, temp;
		sqrt(lamda.t(), lamda_sqrt);
		divide(1,lamda_sqrt,lamda_sqrt);
		temp = repeat(lamda_sqrt,eig_faces.rows,1);
		multiply(eig_faces,temp,eig_faces);

		// Calculate the weights
		Mat weights = eig_faces.t()*facesMat;

		SaveFaceData((m+1), weights, train_index, test_index, classification);

		cout<< "Completed M = "<< (m+1) <<endl;
	}

	cout<<"Finished JackKnife"<<endl;

}



void RunClassify(int num_coef)
{
	vector<double> actrain, actest;
	int num = 1;

	for (int n = 1; n <= 10; n++)
	{

		// Load dat file containing coefficients of each image
		Mat weights;
		vector<int>train_index;
		vector<int>test_index;
		vector<double> classification;

		// Load the face data for set n
		LoadFaceData(n, weights, train_index, test_index, classification);

		// Create test data
		Mat test_data(num_coef, test_index.size(), DataType<double>::type);
		Mat test_group(1,test_index.size(), DataType<double>::type);

		for (int x = 0; x < num_coef; x++)
		{
			int y = 0;
			for (int i = 0; i < test_index.size(); i++)
			{
				test_data.at<double>(x,y) = weights.at<double>(x,test_index[i]);
				y++;

				if(x == 0)
					test_group.at<double>(0,i) = classification[test_index[i]];
			}
		}


		// Create train data
		Mat train_data(num_coef, train_index.size(), DataType<double>::type);
		Mat train_group(1,train_index.size(), DataType<double>::type);
		for (int x = 0; x < num_coef; x++)
		{
			int y = 0;
			for (int i = 0; i < train_index.size(); i++)
			{
				train_data.at<double>(x,y) = weights.at<double>(x,train_index[i]);
				y++;

				if(x == 0)
					train_group.at<double>(0,i) = classification[train_index[i]];
			}
		}

		/* Execute algorithm */
			Mat test_class, train_class;
			//LinearDiscriminant(train_group, train_data, test_data, test_class, train_class);
			//SupportVector(train_group, train_data, test_data, test_class, train_class, n);

		/* End of algorithm */

		int train_correct = 0, test_correct = 0;
		for(int i = 0; i < train_group.cols; i++)
		{
			if(train_group.at<double>(0,i) == train_class.at<double>(0,i))
				train_correct++;
		}

		for(int i = 0; i < test_group.cols; i++)
		{
			if(test_group.at<double>(0,i) == test_class.at<double>(0,i))
				test_correct++;
			//else
			//	cout<<"i = "<<i<<"\t"<<test_group.at<double>(0,i)<<"\t"<<test_class.at<double>(0,i)<<endl;
		}

		actrain.push_back(double(train_correct)/double(train_group.cols));
		actest.push_back(double(test_correct)/double(test_group.cols));

		cout<<"n = "<<n<<" Correct: "<<actrain[actrain.size()-1]<<" "<<actest[actest.size()-1]<<endl;
	}

	double train_acc = VectorMean(actrain);
	double test_acc =  VectorMean(actest);
	cout<<"Average <Training, Testing>: < "<<train_acc<<",\t"<<test_acc<<" >"<<endl;

}
