/*
 * detector.h
 *
 *  Created on: Mar 5, 2012
 *      Author: azfar
 */

#ifndef DETECTOR_H_
#define DETECTOR_H_

#include "highgui.h"
#include "cv.h"
#include "ml.h"
#include "cxcore.h"
#include "cxtypes.h"
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <fstream>
#include <ctime>
#include "dirent.h"
#include "detector.h"

using namespace std;
using namespace cv;

class detector {
public:
	detector();
	detector(string t);
	virtual ~detector();

	/*
	 * This method loads the face images from the dirname directory. The directory must contain only images and two
	 * folders: Positive and Negative. The result of the method is facesMatND which is a three dimensional array that
	 * contains intesity image information in the format of [x-pixel][y-pixel][image] of type double
	 *
	 * The inputs are:
	 * 		dirname:	Directory containing Positive and Negative images
	 * 		size:		Size of the images to be reduced to (images set as size x size)
	 * 		facesMatND:	Three dimensional array containing pixel information about each image
	 * 		Num_Pos:	Number of positive images
	 * 		Num_Neg:	Number of negative images
	 * 		detectFace:	Flag used to turn on face detection algorithm to focus only on the face
	 */
	void LoadFaces (string dirname, int size, MatND &facesMatND, int Num_Pos, int Num_Neg, bool detectFace = false);

	/*
	 * This method is used to classify images that are stored in facesMatND. The results are stored into a matrix
	 * called predict_class. Currently, DCT and SVM are loaded from the Trained/ folder and used to classify. TrainDCT
	 * must be ran before this can be used
	 *
	 * The inputs are:
	 * 		facesMatND:		Three dimensional array containing pixel information about each image
	 * 		predict_class:	This matrix contains a list of the classification in the order that they are in facesMatND
	 */
	void Classify(MatND &facesMatND, Mat &predict_class);

	/*
	 * This method trains the classifier using DCT (Discrete Cosine Transform) and SVM (Support Vector Machines).
	 * The method takes in the facesMatND and the truth vector, trains the SVM, and stores the best SVM into the folder
	 * Trained/Trained_SVM.xml
	 *
	 * The inputs are:
	 * 		size:			Size of the images (images set as size x size)
	 * 		facesMatND:		Three dimensional array containing pixel information about each image
	 * 		classification:	Vector containing the ground truth (1 = Positive, 2 = Negative)
	 * 		Num_Pos:		Number of positive images
	 * 		Num_Neg:		Number of negative images
	 * 		N:				Number of coefficients used in DCT. 310 usually used
	 * 		isTrain:		Flag to indicate if a new SVM needs to be used or the existing one can be used. Used more for retesting
	 */
	void TrainDCT(int size, MatND &facesMatND, vector<double> classification, int NUM_POS, int NUM_NEG, int N, bool isTrain);

	/*
	 * This method works in the same fashion as TrainDCT except it outputs additional statistics including
	 * True Positive, True Negative, False Positive, False Negative. This method is used for ROC curve generation.
	 * The theta value is varied between -2 and 2 and various data is printed to analyze the performance
	 *
	 * Refer the TrainDCT for details
	 */
	void TrainDCT_Stats(int size, MatND &facesMatND, vector<double> classification, int NUM_POS, int NUM_NEG, int N, bool isTrain);
};

#endif /* DETECTOR_H_ */
