/*
 * This code is a sample of how to use the detector library.
 */
#include "highgui.h"
#include "cv.h"
#include "ml.h"
#include "cxcore.h"
#include "cxtypes.h"
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <fstream>
#include <ctime>
#include "dirent.h"
#include "detector.h"

using namespace std;
using namespace cv;

const int NUM_MALE = 672;		// Number of Male images
const int NUM_FEMALE = 479;		// Number of Female images

const bool TRAIN = false;		// Flag to distiguish between training and testing


int main()
{

	int size = 100;			// Pixel size to downsample images to
	vector<double> gender;	// Vector that provides the gender indices relating to facesMatND

	int faces_size[] = {size,size,NUM_MALE+NUM_FEMALE};

	// facesMatND is structured as [x-pixel][y-pixel][image]
	// It is of type double
	MatND facesMatND;
	facesMatND.create(3,faces_size,DataType<double>::type);

	// Sets up the ground truth
	// Male = 1, Female = 2
	for(int i = 0; i<NUM_MALE; i++)
		gender.push_back(1);

	for(int i = 0; i<NUM_FEMALE; i++)
		gender.push_back(2);

	detector D;		// Creates a new instance of Detector

	// If training, faces are loaded and the learner is trained
	if (TRAIN == true)
	{
		// Need to give the relative path to the images used for training
		// This method will read in all the images and return them in the facesMatND
		D.LoadFaces("Data/Train",size, facesMatND, NUM_MALE, NUM_FEMALE);

		// This method will train the classifier
		D.TrainDCT(size, facesMatND, gender, NUM_MALE, NUM_FEMALE, 310,true);
	}
	else
	{
		// This method will read in all the images and return them in the facesMatND
		D.LoadFaces("Data/Train",size, facesMatND, NUM_MALE, NUM_FEMALE);

		// Classify method will take in facesMatND and return a matrix of the predicted outcome in the
		// order they were read
		Mat predict_class;
		D.Classify(facesMatND, predict_class);

		// Determine the accuracy
		double accuracy = 0.0;
		for (int i = 0; i<predict_class.cols; i++)
		{
			if(predict_class.at<double>(0,i) == gender[i])
				accuracy++;
		}

		cout<< "Accuracy = "<<accuracy/double(predict_class.cols)<<endl;

	}


}
